@@ -1,6 +1,7 @@
 # -*- coding: utf-8 -*-
 """Aliases for the xonsh shell."""
 import os
+import re
 import sys
 import inspect
 import argparse
@@ -38,6 +39,11 @@ import xonsh.history.main as xhm
 import xonsh.xoreutils.which as xxw
 
 
+@lazyobject
+def SUB_EXEC_ALIAS_RE():
+    return re.compile(r"@\(|\$\(|!\(|\$\[|!\[")
+
+
 class Aliases(cabc.MutableMapping):
     """Represents a location to hold and look up aliases."""
 
@@ -78,9 +84,18 @@ class Aliases(cabc.MutableMapping):
         if callable(value):
             if acc_args:  # Partial application
 
-                def _alias(args, stdin=None):
+                def _alias(
+                    args, stdin=None, stdout=None, stderr=None, spec=None, stack=None
+                ):
                     args = list(acc_args) + args
-                    return value(args, stdin=stdin)
+                    return value(
+                        args,
+                        stdin=stdin,
+                        stdout=stdout,
+                        stderr=stderr,
+                        spec=spec,
+                        stack=stack,
+                    )
 
                 return _alias
             else:
@@ -121,13 +136,16 @@ class Aliases(cabc.MutableMapping):
 
     def __setitem__(self, key, val):
         if isinstance(val, str):
-            if isexpression(val):
+            f = "<exec-alias:" + key + ">"
+            if SUB_EXEC_ALIAS_RE.search(val) is not None:
+                # We have a sub-command, e.g. $(cmd), to evaluate
+                self._raw[key] = ExecAlias(val, filename=f)
+            elif isexpression(val):
                 # expansion substitution
-                lexer = __xonsh__.execer.parser.lexer
+                lexer = builtins.__xonsh__.execer.parser.lexer
                 self._raw[key] = list(map(strip_simple_quotes, lexer.split(val)))
             else:
                 # need to exec alias
-                f = "<exec-alias:" + key + ">"
                 self._raw[key] = ExecAlias(val, filename=f)
         else:
             self._raw[key] = val
