@@ -3,7 +3,7 @@
 a command will be able to be run in the background.
 
 A background predictor is a function that accepts a single argument list
-and returns whethere or not the process can be run in the background (returns
+and returns whether or not the process can be run in the background (returns
 True) or must be run the foreground (returns False).
 """
 import os
@@ -39,7 +39,7 @@ class CommandsCache(cabc.Mapping):
     def __iter__(self):
         for cmd, (path, is_alias) in self.all_commands.items():
             if ON_WINDOWS and path is not None:
-                # All comand keys are stored in uppercase on Windows.
+                # All command keys are stored in uppercase on Windows.
                 # This ensures the original command name is returned.
                 cmd = pathbasename(path)
             yield cmd
@@ -61,31 +61,37 @@ class CommandsCache(cabc.Mapping):
          name on Windows as a list, conserving the ordering in `PATHEXT`.
          Returns a list as `name` being the only item in it on other platforms."""
         if ON_WINDOWS:
-            pathext = builtins.__xonsh_env__.get('PATHEXT')
+            pathext = builtins.__xonsh__.env.get("PATHEXT", [])
             name = name.upper()
-            return [
-                name + ext
-                for ext in ([''] + pathext)
-            ]
+            return [name + ext for ext in ([""] + pathext)]
         else:
             return [name]
 
+    @staticmethod
+    def remove_dups(p):
+        ret = list()
+        for e in p:
+            if e not in ret:
+                ret.append(e)
+        return ret
+
     @property
     def all_commands(self):
-        paths = builtins.__xonsh_env__.get('PATH', [])
-        pathset = frozenset(x for x in paths if os.path.isdir(x))
+        paths = builtins.__xonsh__.env.get("PATH", [])
+        paths = CommandsCache.remove_dups(paths)
+        path_immut = tuple(x for x in paths if os.path.isdir(x))
         # did PATH change?
-        path_hash = hash(pathset)
+        path_hash = hash(path_immut)
         cache_valid = path_hash == self._path_checksum
         self._path_checksum = path_hash
         # did aliases change?
-        alss = getattr(builtins, 'aliases', dict())
+        alss = getattr(builtins, "aliases", dict())
         al_hash = hash(frozenset(alss))
         cache_valid = cache_valid and al_hash == self._alias_checksum
         self._alias_checksum = al_hash
         # did the contents of any directory in PATH change?
         max_mtime = 0
-        for path in pathset:
+        for path in path_immut:
             mtime = os.stat(path).st_mtime
             if mtime > max_mtime:
                 max_mtime = mtime
@@ -94,7 +100,7 @@ class CommandsCache(cabc.Mapping):
         if cache_valid:
             return self._cmds_cache
         allcmds = {}
-        for path in reversed(paths):
+        for path in reversed(path_immut):
             # iterate backwards so that entries at the front of PATH overwrite
             # entries at the back.
             for cmd in executables_in(path):
@@ -172,21 +178,41 @@ class CommandsCache(cabc.Mapping):
         if ON_WINDOWS:
             # Windows users expect to be able to execute files in the same
             # directory without `./`
-            local_bin = next((fn for fn in possibilities if os.path.isfile(fn)),
-                             None)
+            local_bin = next((fn for fn in possibilities if os.path.isfile(fn)), None)
             if local_bin:
                 return os.path.abspath(local_bin)
-        cached = next((cmd for cmd in possibilities if cmd in self._cmds_cache),
-                      None)
+        cached = next((cmd for cmd in possibilities if cmd in self._cmds_cache), None)
         if cached:
             (path, alias) = self._cmds_cache[cached]
-            if not alias or ignore_alias:
-                return path
-            else:
+            ispure = path == pathbasename(path)
+            if alias and ignore_alias and ispure:
+                # pure alias, which we are ignoring
                 return None
+            else:
+                return path
         elif os.path.isfile(name) and name != pathbasename(name):
             return name
 
+    def is_only_functional_alias(self, name):
+        """Returns whether or not a command is only a functional alias, and has
+        no underlying executable. For example, the "cd" command is only available
+        as a functional alias.
+        """
+        _ = self.all_commands
+        return self.lazy_is_only_functional_alias(name)
+
+    def lazy_is_only_functional_alias(self, name):
+        """Returns whether or not a command is only a functional alias, and has
+        no underlying executable. For example, the "cd" command is only available
+        as a functional alias. This search is performed lazily.
+        """
+        val = self._cmds_cache.get(name, None)
+        if val is None:
+            return False
+        return (
+            val == (name, True) and self.locate_binary(name, ignore_alias=True) is None
+        )
+
     def predict_threadable(self, cmd):
         """Predicts whether a command list is able to be run on a background
         thread, rather than the main thread.
@@ -216,14 +242,14 @@ class CommandsCache(cabc.Mapping):
 
     def default_predictor(self, name, cmd0):
         if ON_POSIX:
-            return self.default_predictor_readbin(name, cmd0,
-                                                  timeout=0.1,
-                                                  failure=predict_true)
+            return self.default_predictor_readbin(
+                name, cmd0, timeout=0.1, failure=predict_true
+            )
         else:
             return predict_true
 
     def default_predictor_readbin(self, name, cmd0, timeout, failure):
-        """Make a defautt predictor by
+        """Make a default predictor by
         analyzing the content of the binary. Should only works on POSIX.
         Return failure if the analysis fails.
         """
@@ -242,11 +268,12 @@ class CommandsCache(cabc.Mapping):
             return failure  # opening error
 
         search_for = {
-            (b'ncurses',): [False, ],
-            (b'isatty', b'tcgetattr', b'tcsetattr'): [False, False, False],
+            (b"ncurses",): [False],
+            (b"libgpm",): [False],
+            (b"isatty", b"tcgetattr", b"tcsetattr"): [False, False, False],
         }
         tstart = time.time()
-        block = b''
+        block = b""
         while time.time() < tstart + timeout:
             previous_block = block
             try:
@@ -290,14 +317,14 @@ def predict_false(args):
 
 @lazyobject
 def SHELL_PREDICTOR_PARSER():
-    p = argparse.ArgumentParser('shell', add_help=False)
-    p.add_argument('-c', nargs='?', default=None)
-    p.add_argument('filename', nargs='?', default=None)
+    p = argparse.ArgumentParser("shell", add_help=False)
+    p.add_argument("-c", nargs="?", default=None)
+    p.add_argument("filename", nargs="?", default=None)
     return p
 
 
 def predict_shell(args):
-    """Precict the backgroundability of the normal shell interface, which
+    """Predict the backgroundability of the normal shell interface, which
     comes down to whether it is being run in subproc mode.
     """
     ns, _ = SHELL_PREDICTOR_PARSER.parse_known_args(args)
@@ -310,16 +337,16 @@ def predict_shell(args):
 
 @lazyobject
 def HELP_VER_PREDICTOR_PARSER():
-    p = argparse.ArgumentParser('cmd', add_help=False)
-    p.add_argument('-h', '--help', dest='help',
-                   action='store_true', default=None)
-    p.add_argument('-v', '-V', '--version', dest='version',
-                   action='store_true', default=None)
+    p = argparse.ArgumentParser("cmd", add_help=False)
+    p.add_argument("-h", "--help", dest="help", action="store_true", default=None)
+    p.add_argument(
+        "-v", "-V", "--version", dest="version", action="store_true", default=None
+    )
     return p
 
 
 def predict_help_ver(args):
-    """Precict the backgroundability of commands that have help & version
+    """Predict the backgroundability of commands that have help & version
     switches: -h, --help, -v, -V, --version. If either of these options is
     present, the command is assumed to print to stdout normally and is therefore
     threadable. Otherwise, the command is assumed to not be threadable.
@@ -331,53 +358,84 @@ def predict_help_ver(args):
     return pred
 
 
+@lazyobject
+def HG_PREDICTOR_PARSER():
+    p = argparse.ArgumentParser("hg", add_help=False)
+    p.add_argument("command")
+    p.add_argument(
+        "-i", "--interactive", action="store_true", default=False, dest="interactive"
+    )
+    return p
+
+
+def predict_hg(args):
+    """Predict if mercurial is about to be run in interactive mode.
+    If it is interactive, predict False. If it isn't, predict True.
+    Also predict False for certain commands, such as split.
+    """
+    ns, _ = HG_PREDICTOR_PARSER.parse_known_args(args)
+    if ns.command == "split":
+        return False
+    else:
+        return not ns.interactive
+
+
 def default_threadable_predictors():
     """Generates a new defaultdict for known threadable predictors.
     The default is to predict true.
     """
     # alphabetical, for what it is worth.
     predictors = {
-        'bash': predict_shell,
-        'csh': predict_shell,
-        'clear': predict_false,
-        'cls': predict_false,
-        'cmd': predict_shell,
-        'ex': predict_false,
-        'fish': predict_shell,
-        'gvim': predict_help_ver,
-        'htop': predict_help_ver,
-        'ksh': predict_shell,
-        'less': predict_help_ver,
-        'man': predict_help_ver,
-        'more': predict_help_ver,
-        'mvim': predict_help_ver,
-        'mutt': predict_help_ver,
-        'nano': predict_help_ver,
-        'nvim': predict_false,
-        'ponysay': predict_help_ver,
-        'psql': predict_false,
-        'python': predict_shell,
-        'python2': predict_shell,
-        'python3': predict_shell,
-        'ranger': predict_help_ver,
-        'rview': predict_false,
-        'rvim': predict_false,
-        'scp': predict_false,
-        'sh': predict_shell,
-        'ssh': predict_false,
-        'startx': predict_false,
-        'sudo': predict_help_ver,
-        'tcsh': predict_shell,
-        'telnet': predict_false,
-        'top': predict_help_ver,
-        'vi': predict_false,
-        'view': predict_false,
-        'vim': predict_false,
-        'vimpager': predict_help_ver,
-        'weechat': predict_help_ver,
-        'xo': predict_help_ver,
-        'xonsh': predict_shell,
-        'xon.sh': predict_shell,
-        'zsh': predict_shell,
+        "aurman": predict_false,
+        "bash": predict_shell,
+        "csh": predict_shell,
+        "clear": predict_false,
+        "cls": predict_false,
+        "cmd": predict_shell,
+        "cryptop": predict_false,
+        "curl": predict_true,
+        "ex": predict_false,
+        "emacsclient": predict_false,
+        "fish": predict_shell,
+        "gvim": predict_help_ver,
+        "hg": predict_hg,
+        "htop": predict_help_ver,
+        "ipython": predict_shell,
+        "ksh": predict_shell,
+        "less": predict_help_ver,
+        "ls": predict_true,
+        "man": predict_help_ver,
+        "more": predict_help_ver,
+        "mvim": predict_help_ver,
+        "mutt": predict_help_ver,
+        "nano": predict_help_ver,
+        "nvim": predict_false,
+        "ponysay": predict_help_ver,
+        "psql": predict_false,
+        "python": predict_shell,
+        "python2": predict_shell,
+        "python3": predict_shell,
+        "repo": predict_help_ver,
+        "ranger": predict_help_ver,
+        "rview": predict_false,
+        "rvim": predict_false,
+        "scp": predict_false,
+        "sh": predict_shell,
+        "ssh": predict_false,
+        "startx": predict_false,
+        "sudo": predict_help_ver,
+        "tcsh": predict_shell,
+        "telnet": predict_false,
+        "top": predict_help_ver,
+        "vi": predict_false,
+        "view": predict_false,
+        "vim": predict_false,
+        "vimpager": predict_help_ver,
+        "weechat": predict_help_ver,
+        "xclip": predict_help_ver,
+        "xo": predict_help_ver,
+        "xonsh": predict_shell,
+        "xon.sh": predict_shell,
+        "zsh": predict_shell,
     }
     return predictors
